#include "OgreDepthShadowReceiverRenderTechnique.h"
#include "OgreTechniqueGroup.h"
#include "OgreIlluminationManager.h"
#include "OgreDepthShadowMapRenderingRun.h"

OgreDepthShadowReceiverRenderTechnique::OgreDepthShadowReceiverRenderTechnique(
												int maxlights,
												String shadowVertexProgram,
												String shadowFragmentProgram,
												String WorldViewProjParamName,
												String WorldParamName,
												bool setLightViewMatrix,
												bool setLightViewProjMatrix,
												bool setLightProjFarPlane,
												String lightViewProjParamName,
												String lightViewParamName,
												String lightFarPlaneParamName,
												SceneBlendFactor passBlendingSRC,
												SceneBlendFactor passBlendingDEST,
												bool createNewPasses,
												int startTextureUnitID,							
												Pass* pass,
												OgreRenderable* parentRenderable,
												OgreTechniqueGroup* parentTechniqueGroup)
							:RenderTechnique( parentRenderable, parentTechniqueGroup),
							OgreRenderTechnique(pass, parentRenderable, parentTechniqueGroup),
							DepthShadowReceiverRenderTechnique(parentRenderable, parentTechniqueGroup)
{
	this->passBlendingSRC = passBlendingSRC;
	this->passBlendingDEST = passBlendingDEST;
	this->maxlights = maxlights;
	this->shadowVertexProgram = shadowVertexProgram ;
	this->shadowFragmentProgram = shadowFragmentProgram;
	this->setLightViewMatrix = setLightViewMatrix;
	this->setLightViewProjMatrix = setLightViewProjMatrix;
	this->setLightProjFarPlane = setLightProjFarPlane;
	this->lightViewProjParamName = lightViewProjParamName;
	this->lightViewParamName = lightViewParamName;
	this->lightFarPlaneParamName = lightFarPlaneParamName;
	this->WorldParamName = WorldParamName;
	this->WorldViewProjParamName = WorldViewProjParamName;
	this->createNewPasses = createNewPasses;
	this->startTextureUnitID = startTextureUnitID;
							
	if(createNewPasses)
	{
		//insert new passes
		Ogre::Technique* techn = pass->getParent();
		Technique::PassIterator it = techn->getPassIterator();
		
		int index = 0;
		while(it.hasMoreElements())
		{
			if( it.getNext() == pass)
				break;
			index++;
			it.moveNext();
		}
		
		index++;
		for(int i = 0; i < maxlights; i++)
		{
			int lastpass = techn->getNumPasses();
			Pass* newpass = techn->createPass();
			passes.push_back(newpass);
			
			TextureUnitState* st = newpass->createTextureUnitState();		
			st->setTextureFiltering(TFO_NONE);
			st->setTextureAddressingMode(TextureUnitState::TAM_BORDER);
			st->setTextureBorderColour(ColourValue::White);

			newpass->setSceneBlending(passBlendingSRC, passBlendingDEST);
			
			newpass->setDepthBias(1);
			techn->movePass(lastpass, index);			
		}
	}	
}

OgreDepthShadowReceiverRenderTechnique::~OgreDepthShadowReceiverRenderTechnique()
{

}

void OgreDepthShadowReceiverRenderTechnique::update(unsigned long frameNum)
{
	LightList lights;
	Root::getSingleton()._getCurrentSceneManager()->_populateLightList(
						((OgreSharedRuns*) sharedRuns)->getRootPosition(),
							100000.0,
							lights);

	//fill passes
	unsigned int l = 0;
	for(unsigned int i = 0; i < maxlights; i++)
	{
		while(lights.size() != l && !lights.at(l)->getCastShadows())
		 l++;

		if(lights.size() > l)
		{
			
				//create run if not exists
				OgreIlluminationManager::getSingleton().createPerLightRun(lights.at(l)->getName(),
																			ILLUMRUN_DEPTH_SHADOWMAP);
				//update light depth map
				OgreIlluminationManager::getSingleton().updatePerLightRun(lights.at(l)->getName(),
																			ILLUMRUN_DEPTH_SHADOWMAP,
																			frameNum);
				
				//set texture to pass
				OgreDepthShadowMapRenderingRun* depthRun = 
					(OgreDepthShadowMapRenderingRun*) OgreIlluminationManager::getSingleton()
														.getPerLightRun(lights.at(l)->getName(), 
														ILLUMRUN_DEPTH_SHADOWMAP)->asOgreRenderingRun();
				
				Pass* passToSet = this->pass;
				int textureUnit = startTextureUnitID + i;
				if(createNewPasses)
				{
					passToSet = passes.at(i);
					passToSet->setActive(true);
					textureUnit = 0;
					passes.at(i)->setActive(true);
				}
				passToSet->getTextureUnitState(textureUnit)->setTextureName(
					depthRun->getDepthMapTextureName());
			
				if(createNewPasses)
				{
					//TODO: ask programs from illummanager
					passes.at(i)->setVertexProgram(shadowVertexProgram);
					passes.at(i)->setFragmentProgram(shadowFragmentProgram);
				}
		
				GpuProgramParametersSharedPtr fpParams = passToSet->getFragmentProgramParameters();
				GpuProgramParametersSharedPtr vpParams = passToSet->getVertexProgramParameters();

				String LightViewProjParamName = lightViewProjParamName;
				String LightViewParamName = lightViewParamName;
				String LightFarPlaneParamName = lightFarPlaneParamName;
				if(!createNewPasses)
				{
					LightViewProjParamName += StringConverter::toString(i+1);
					LightViewParamName += StringConverter::toString(i+1);
					LightFarPlaneParamName += StringConverter::toString(i+1);
				}
				else
				{
					vpParams->setNamedAutoConstant(WorldViewProjParamName,
										GpuProgramParameters::ACT_WORLDVIEWPROJ_MATRIX);	
					vpParams->setNamedAutoConstant(WorldParamName,
										GpuProgramParameters::ACT_WORLD_MATRIX);
				}
		
				if(setLightViewProjMatrix)
				{
					if(createNewPasses)
					vpParams->setNamedConstant(LightViewProjParamName, depthRun->getLightViewProjMatrix());
					else
					fpParams->setNamedConstant(LightViewProjParamName, depthRun->getLightViewProjMatrix());
				}
				if(setLightViewMatrix)
				{
					if(createNewPasses)
					vpParams->setNamedConstant(LightViewParamName, depthRun->getLightViewMatrix());
					else
					fpParams->setNamedConstant(LightViewParamName, depthRun->getLightViewMatrix());
				}
				if(setLightProjFarPlane)
				{
					if(createNewPasses)
					fpParams->setNamedConstant(LightFarPlaneParamName, depthRun->getLightFarPlane());
					else
					fpParams->setNamedConstant(LightFarPlaneParamName, depthRun->getLightFarPlane());
				}
		}	
		else if(createNewPasses)
			passes.at(i)->setActive(false);

		l++;
	}
}


///Technique parsers
namespace DepthShadowReceiverParsers
{
	void parseMaxLights(String& params, RenderTechniqueFactory* factory)
	{
		OgreDepthShadowReceiverRenderTechniqueFactory* f = (OgreDepthShadowReceiverRenderTechniqueFactory*) factory;
		f->maxlights =  StringConverter::parseInt(params);
	}

	void parseVertexProgram(String& params, RenderTechniqueFactory* factory)
	{
		OgreDepthShadowReceiverRenderTechniqueFactory* f = (OgreDepthShadowReceiverRenderTechniqueFactory*) factory;
		f->shadowVertexProgram =  params;
	}

	void parseFragmentProgram(String& params, RenderTechniqueFactory* factory)
	{
		OgreDepthShadowReceiverRenderTechniqueFactory* f = (OgreDepthShadowReceiverRenderTechniqueFactory*) factory;
		f->shadowFragmentProgram =  params;
	}

	void parseSetLightViewProj(String& params, RenderTechniqueFactory* factory)
	{
		OgreDepthShadowReceiverRenderTechniqueFactory* f = (OgreDepthShadowReceiverRenderTechniqueFactory*) factory;
		StringVector vecparams = StringUtil::split(params, " \t");
		f->setLightViewProjMatrix = StringConverter::parseBool(vecparams[0]);
		if(f->setLightViewProjMatrix && vecparams.size() > 1)
			f->lightViewProjParamName =  vecparams[1];
	}

	void parseSetLightView(String& params, RenderTechniqueFactory* factory)
	{
		OgreDepthShadowReceiverRenderTechniqueFactory* f = (OgreDepthShadowReceiverRenderTechniqueFactory*) factory;
		StringVector vecparams = StringUtil::split(params, " \t");
		f->setLightViewMatrix = StringConverter::parseBool(vecparams[0]);
		if(f->setLightViewMatrix && vecparams.size() > 1)
			f->lightViewParamName =  vecparams[1];
	}

	void parseSetLightFarPlane(String& params, RenderTechniqueFactory* factory)
	{
		OgreDepthShadowReceiverRenderTechniqueFactory* f = (OgreDepthShadowReceiverRenderTechniqueFactory*) factory;
		StringVector vecparams = StringUtil::split(params, " \t");
		f->setLightProjFarPlane = StringConverter::parseBool(vecparams[0]);
		if(f->setLightProjFarPlane && vecparams.size() > 1)
			f->lightFarPlaneParamName =  vecparams[1];
	}

	void parseLightProjParamName(String& params, RenderTechniqueFactory* factory)
	{
		OgreDepthShadowReceiverRenderTechniqueFactory* f = (OgreDepthShadowReceiverRenderTechniqueFactory*) factory;
		f->lightViewProjParamName =  params;
	}

	void parseLightViewParamName(String& params, RenderTechniqueFactory* factory)
	{
		OgreDepthShadowReceiverRenderTechniqueFactory* f = (OgreDepthShadowReceiverRenderTechniqueFactory*) factory;
		f->lightViewParamName =  params;
	}

	void parseWorldViewProjParamName(String& params, RenderTechniqueFactory* factory)
	{
		OgreDepthShadowReceiverRenderTechniqueFactory* f = (OgreDepthShadowReceiverRenderTechniqueFactory*) factory;
		f->WorldViewProjParamName =  params;
	}

	void parseWorldParamName(String& params, RenderTechniqueFactory* factory)
	{
		OgreDepthShadowReceiverRenderTechniqueFactory* f = (OgreDepthShadowReceiverRenderTechniqueFactory*) factory;
		f->WorldParamName =  params;
	}

	void parseLightFarPlaneParamName(String& params, RenderTechniqueFactory* factory)
	{
		OgreDepthShadowReceiverRenderTechniqueFactory* f = (OgreDepthShadowReceiverRenderTechniqueFactory*) factory;
		f->lightFarPlaneParamName =  params;
	}

	SceneBlendFactor convertBlendFactor(const String& param)
    {
        if (param == "one")
            return SBF_ONE;
        else if (param == "zero")
            return SBF_ZERO;
        else if (param == "dest_colour")
            return SBF_DEST_COLOUR;
        else if (param == "src_colour")
            return SBF_SOURCE_COLOUR;
        else if (param == "one_minus_dest_colour")
            return SBF_ONE_MINUS_DEST_COLOUR;
        else if (param == "one_minus_src_colour")
            return SBF_ONE_MINUS_SOURCE_COLOUR;
        else if (param == "dest_alpha")
            return SBF_DEST_ALPHA;
        else if (param == "src_alpha")
            return SBF_SOURCE_ALPHA;
        else if (param == "one_minus_dest_alpha")
            return SBF_ONE_MINUS_DEST_ALPHA;
        else if (param == "one_minus_src_alpha")
            return SBF_ONE_MINUS_SOURCE_ALPHA;       
    }

	void parsePassBlending(String& params, RenderTechniqueFactory* factory)
	{
		OgreDepthShadowReceiverRenderTechniqueFactory* f = (OgreDepthShadowReceiverRenderTechniqueFactory*) factory;
		StringVector vecparams = StringUtil::split(params, " \t");
		if(vecparams.size() == 1)
		{
			if (vecparams[0] == "none")
			{
                f->passBlendingSRC =  SBF_ONE;
				f->passBlendingDEST = SBF_ZERO;
			}
			if (vecparams[0] == "add")
			{
                f->passBlendingSRC =  SBF_ONE;
				f->passBlendingDEST = SBF_ONE;
			}
            else if (vecparams[0] == "modulate")
            {
				f->passBlendingSRC =  SBF_DEST_COLOUR;
				f->passBlendingDEST = SBF_ZERO;
			}
			else if (vecparams[0] == "colour_blend")
			{
				f->passBlendingSRC =  SBF_SOURCE_COLOUR;
				f->passBlendingDEST = SBF_ONE_MINUS_SOURCE_COLOUR;
			}
            else if (vecparams[0] == "alpha_blend")
            {
                f->passBlendingSRC =  SBF_SOURCE_ALPHA;
				f->passBlendingDEST = SBF_ONE_MINUS_SOURCE_ALPHA;
			}            
		}
		else if (vecparams.size() == 2)
		{
		  f->passBlendingSRC =  convertBlendFactor(vecparams[0]);
		  f->passBlendingDEST = convertBlendFactor(vecparams[1]) ;
		}		
	}

	void parseCreateNewPasses(String& params, RenderTechniqueFactory* factory)
	{
		OgreDepthShadowReceiverRenderTechniqueFactory* f = (OgreDepthShadowReceiverRenderTechniqueFactory*) factory;
		f->createNewPasses =  StringConverter::parseBool(params);
	}
	
	void parseStartTexID(String& params, RenderTechniqueFactory* factory)
	{
		OgreDepthShadowReceiverRenderTechniqueFactory* f = (OgreDepthShadowReceiverRenderTechniqueFactory*) factory;
		f->startTextureUnitID =  StringConverter::parseInt(params);
	}
}

OgreDepthShadowReceiverRenderTechniqueFactory::OgreDepthShadowReceiverRenderTechniqueFactory()
{
	typeName = "DepthShadowReceiver";

	using namespace DepthShadowReceiverParsers;
	//register parsers
	this->attributeParsers.insert(AttribParserList::value_type("max_light_count", (ILLUM_ATTRIBUTE_PARSER) parseMaxLights));
	this->attributeParsers.insert(AttribParserList::value_type("vertex_program_name", (ILLUM_ATTRIBUTE_PARSER) parseVertexProgram));
	this->attributeParsers.insert(AttribParserList::value_type("fragment_program_name", (ILLUM_ATTRIBUTE_PARSER) parseFragmentProgram));
	this->attributeParsers.insert(AttribParserList::value_type("set_light_viewproj", (ILLUM_ATTRIBUTE_PARSER) parseSetLightViewProj));
	this->attributeParsers.insert(AttribParserList::value_type("set_light_view", (ILLUM_ATTRIBUTE_PARSER) parseSetLightView));
	this->attributeParsers.insert(AttribParserList::value_type("set_light_farplane", (ILLUM_ATTRIBUTE_PARSER) parseSetLightFarPlane));
	this->attributeParsers.insert(AttribParserList::value_type("light_viewproj_param_name", (ILLUM_ATTRIBUTE_PARSER) parseLightProjParamName));
	this->attributeParsers.insert(AttribParserList::value_type("light_view_param_name", (ILLUM_ATTRIBUTE_PARSER) parseLightViewParamName));
	this->attributeParsers.insert(AttribParserList::value_type("world_view_proj_param_name", (ILLUM_ATTRIBUTE_PARSER) parseWorldViewProjParamName));
	this->attributeParsers.insert(AttribParserList::value_type("world_param_name", (ILLUM_ATTRIBUTE_PARSER) parseWorldParamName));
	this->attributeParsers.insert(AttribParserList::value_type("light_farplane_param_name", (ILLUM_ATTRIBUTE_PARSER) parseLightFarPlaneParamName));
	this->attributeParsers.insert(AttribParserList::value_type("pass_blending", (ILLUM_ATTRIBUTE_PARSER) parsePassBlending));
	this->attributeParsers.insert(AttribParserList::value_type("new_passes", (ILLUM_ATTRIBUTE_PARSER) parseCreateNewPasses));
	this->attributeParsers.insert(AttribParserList::value_type("start_tex_id", (ILLUM_ATTRIBUTE_PARSER) parseStartTexID));	
}

OgreRenderTechnique* OgreDepthShadowReceiverRenderTechniqueFactory::createInstance( 
										IllumTechniqueParams* params,
										Pass* pass,
										OgreRenderable* parentRenderable,
										OgreTechniqueGroup* parentTechniqueGroup)
{	
	//reset parameters
	maxlights = 1;
	shadowVertexProgram = "GTP/Basic/LightCPos_VS";
	shadowFragmentProgram = "GTP/Basic/SM/Dist_PS";
	setLightViewMatrix = true;
	setLightViewProjMatrix = true;
	setLightProjFarPlane = false;
	lightViewProjParamName = "LightViewProj";
	lightViewParamName = "LightView";
	lightFarPlaneParamName = "lightFarPlane";
	WorldViewProjParamName = "WorldViewProj";
	WorldParamName = "World";
	passBlendingSRC = SBF_DEST_COLOUR;
	passBlendingDEST = SBF_ZERO;
	createNewPasses = true;
	startTextureUnitID = 0;
	
	parseParams(params);

	OgreDepthShadowReceiverRenderTechnique* result = new OgreDepthShadowReceiverRenderTechnique(
												maxlights,
												shadowVertexProgram,
												shadowFragmentProgram,
												WorldViewProjParamName,
												WorldParamName,
												setLightViewMatrix ,
												setLightViewProjMatrix,
												setLightProjFarPlane,
												lightViewProjParamName,
												lightViewParamName,
												lightFarPlaneParamName,
												passBlendingSRC,
												passBlendingDEST,
												createNewPasses,
												startTextureUnitID,
												pass,
												parentRenderable,
												parentTechniqueGroup);	

	return result;
}

